"""
GenePattern Notebook extension for Jupyter

Copyright 2015-2021 Regents of the University of California & Broad Institute
"""

from .remote_widgets import GPAuthWidget, GPTaskWidget, GPJobWidget, session, register_session, get_session, display
from nbtools import UIBuilder as GPUIBuilder, UIOutput as GPUIOutput, build_ui, open
from .local_widgets import GPModuleWidget
from .jupyter_extensions import (GenePatternMagic, load_ipython_extension, load_jupyter_server_extension,
                                 _jupyter_server_extension_paths, _jupyter_nbextension_paths)
from .demo_mode import activate_demo_mode, set_demo_jobs


__author__ = 'Thorin Tabor'
__copyright__ = 'Copyright 2014-2021, Regents of the University of California & Broad Institute'
__version__ = '21.12'
__status__ = 'Production/Stable'
__license__ = 'BSD-3-Clause'



import os
from setuptools import setup


__version__ = '21.12'


with open('README.md') as f:
    long_description = f.read()


def get_data_files():
    """
    Get the data files for the package.
    """
    return [
        ('share/jupyter/nbextensions/genepattern', [
            'genepattern/static/index.js',
        ]),
        ('share/jupyter/nbextensions/genepattern/resources',
         ['genepattern/static/resources/' + f for f in os.listdir('genepattern/static/resources')]
         ),
        ('etc/jupyter/nbconfig/notebook.d', ['genepattern.json']),
        # ('share/jupyter/lab/extensions', [
        #     'genepattern/static/index.js',
        #     'genepattern/static/resources',
        # ])
    ]


setup(name='genepattern-notebook',
      packages=['genepattern'],
      version=__version__,
      description='GenePattern Notebook extension for Jupyter',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='BSD',
      author='Thorin Tabor',
      author_email='tmtabor@cloud.ucsd.edu',
      url='https://github.com/genepattern/genepattern-notebook',
      download_url='https://github.com/genepattern/genepattern-notebook/archive/' + __version__ + '.tar.gz',
      keywords=['genepattern', 'genomics', 'bioinformatics', 'ipython', 'jupyter'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Framework :: Jupyter',
      ],
      install_requires=[
          'genepattern-python>=1.4.2',
          'nbtools>=19',
          'notebook>=5.0.0',
          'ipywidgets>=7.0.0',
      ],
      package_data={'genepattern': ['static/index.js', 'static/resources/*']},
      data_files=get_data_files(),
      normalize_version=False,
      )



import warnings

import gp
import IPython.display
from ipywidgets import widgets
from traitlets import Unicode, Integer

"""
Widgets which retrieve results from the GenePattern server

    * GPAuthWidget: GenePattern server authentication
    * GPTaskWidget: Analysis for submission to GenePattern server
    * GPJobWidget: Submitted GenePattern  job
"""


class SessionList:
    """
    Keeps a list of all currently registered GenePattern server sessions
    """
    sessions = []

    def register(self, server, username, password):
        """
        Register a new GenePattern server session for the provided
        server, username and password. Return the session.
        :param server:
        :param username:
        :param password:
        :return:
        """

        # Create the session
        session = gp.GPServer(server, username, password)

        # Validate username if not empty
        valid_username = username != "" and username is not None

        # Validate that the server is not already registered
        index = self._get_index(server)
        new_server = index == -1

        # Add the new session to the list
        if valid_username and new_server:
            self.sessions.append(session)

        # Replace old session is one exists
        if valid_username and not new_server:
            self.sessions[index] = session

        return session

    def get(self, server):
        """
        Returns a registered GPServer object with a matching GenePattern server url or index
        Returns None if no matching result was found
        :param server:
        :return:
        """

        # Handle indexes
        if isinstance(server, int):
            if server >= len(self.sessions):
                return None
            else:
                return self.sessions[server]

        # Handle server URLs
        index = self._get_index(server)
        if index == -1:
            return None
        else:
            return self.sessions[index]

    def clean(self):
        """
        Clear all GenePattern sessions from the sessions list
        :return:
        """
        self.sessions = []

    def _get_index(self, server_url):
        """
        Returns a registered GPServer object with a matching GenePattern server url
        Returns -1 if no matching result was found
        :param server_url:
        :return:
        """
        for i in range(len(self.sessions)):
            session = self.sessions[i]
            if session.url == server_url:
                return i
        return -1


"""
GenePattern Sessions Singleton
"""
session = SessionList()


# Deprecated function calls for backward compatibility
def register_session(server, username, password):
    warnings.warn("register_session will be deprecated in version 1.0, use sessions.register instead", PendingDeprecationWarning)
    return session.register(server, username, password)


def get_session(server):
    warnings.warn("get_session will be deprecated in version 1.0, use sessions.get instead", PendingDeprecationWarning)
    return session.get(server)


class GPAuthWidget(gp.GPResource, widgets.DOMWidget):
    """
    A running or completed job on a GenePattern server.

    Contains methods to get the info of the job, and to wait on a running job by
    polling the server until the job is completed.
    """
    _view_name = Unicode('AuthWidgetView').tag(sync=True)
    _view_module = Unicode('genepattern/authentication').tag(sync=True)

    def __init__(self, uri, **kwargs):
        super(GPAuthWidget, self).__init__(uri)
        widgets.DOMWidget.__init__(self, **kwargs)
        self.info = None

        self.errors = widgets.CallbackDispatcher(accepted_nargs=[0, 1])
        self.on_msg(self._handle_custom_msg)

    def _handle_custom_msg(self, content, **kwargs):
        if 'event' in content and content['event'] == 'error':
            self.errors()
            self.errors(self)


class GPJobWidget(gp.GPResource, widgets.DOMWidget):
    """
    A running or completed job on a GenePattern server.

    Contains methods to get the info of the job, and to wait on a running job by
    polling the server until the job is completed.
    """
    _view_name = Unicode('JobWidgetView').tag(sync=True)
    _view_module = Unicode("genepattern/job").tag(sync=True)
    job_number = Integer(-1).tag(sync=True)

    def __init__(self, job=None, **kwargs):
        # Handle None as placeholder job widget
        super(GPJobWidget, self).__init__(job.uri if hasattr(job, 'uri') else None)
        widgets.DOMWidget.__init__(self, **kwargs)
        if job is not None:
            self.job_number = job.job_number


class GPTaskWidget(gp.GPResource, widgets.DOMWidget):
    """
    A running or completed job on a GenePattern server.

    Contains methods to get the info of the job, and to wait on a running job by
    polling the server until the job is completed.
    """
    _view_name = Unicode('TaskWidgetView').tag(sync=True)
    _view_module = Unicode("genepattern/task").tag(sync=True)
    lsid = Unicode("", sync=True)
    name = Unicode("", sync=True)

    def __init__(self, task, **kwargs):
        super(GPTaskWidget, self).__init__(task.uri)
        widgets.DOMWidget.__init__(self, **kwargs)

        if self.is_uri_lsid(task.uri):
            self.lsid = task.uri
        else:
            self.name = task.uri

    @staticmethod
    def is_uri_lsid(uri):
        if uri.startswith("urn"):
            return True
        else:
            return False


def display(content):
    """
    Display a widget, text or other media in a notebook without the need to import IPython at the top level.

    Also handles wrapping GenePattern Python Library content in widgets.
    :param content:
    :return:
    """
    if isinstance(content, gp.GPServer):
        IPython.display.display(GPAuthWidget(content))
    elif isinstance(content, gp.GPTask):
        IPython.display.display(GPTaskWidget(content))
    elif isinstance(content, gp.GPJob):
        IPython.display.display(GPJobWidget(content))
    else:
        IPython.display.display(content)


from IPython.display import display, Javascript
from json import dumps

demo_jobs = []


"""
    BELOW IS AN EXAMPLE OF HOW TO USE DEMO MODE. SUGGESTED USE: WRAP BEHIND UI BUILDER
    
    import genepattern
    
    genepattern.set_demo_jobs([
        {
            'name': 'Conos.Preprocess',
            'job': 389201,  # Make sure to set the permissions of any demo job to 'public'
            'params': {
                'knn': '40',  # Demo mode should be able to match between int and string, but when in doubt use string
                'perplexity': 50
            }
        }, 
        {
            'name': 'Conos.Cluster',
            'job': 82094,
            'params': {
                'conos_object': 'conos_preprocess_output.rds',  # For file parameters, just list the file name, not URL
                'runleiden': 'True',
                'leiden_resolution': '1.0',
                'runwalktrap': 'True',
                'walktrap_steps': '10',
                'umap_distance': '0.05',
                'umap_spread': '5'
            }
        }
    ])
    
    genepattern.activate_demo_mode()
    display(HTML('<div class="alert alert-success">Demo mode activated</div>'))
"""


def set_demo_jobs(jobs):
    global demo_jobs
    demo_jobs = jobs


def demo_jobs_string():
    global demo_jobs
    return dumps(demo_jobs)


def activate_demo_mode():
    display(Javascript("""
        const demo_jobs = """ + demo_jobs_string() + """;

        const file_special_handling = (uiInput, uiValue) => {
            const is_file_param = $(uiInput).hasClass("nbtools-file");
            if (is_file_param) {
                if (uiValue.constructor === Array) {
                    return uiValue.map(val => {
                        if (val.constructor === File) return val.name;
                        else return val.substring(val.lastIndexOf('/')+1)
                    });
                }
                else {
                    return uiValue.substring(uiValue.lastIndexOf('/')+1);
                }
            }
            else {
                return uiValue;
            }
        }

        const get_actual_values = widget => {
            const values = {};
            const uiParams = widget.element.find(".gp-widget-task-param");
            for (let i = 0; i < uiParams.length; i++) {
                const uiParam = $(uiParams[i]);
                const uiInput = uiParam.find(".gp-widget-task-param-input");
                let uiValue = widget._getInputValue(uiInput);
                uiValue = file_special_handling(uiInput, uiValue);
                const uiName = widget._getParamName(uiParam);
                if (uiValue !== null) {
                    // Wrap value in list if not already wrapped
                    if (uiValue.constructor !== Array) {
                        uiValue = [uiValue];
                    }
                    values[uiName] = uiValue;
                }
            }
            return values;
        }

        const params_match = (job, params) => {
            let matched = true;
            Object.keys(job.params).forEach(key => {
                let value = job.params[key];
                if (value.constructor !== Array) {
                    value = [value];
                }
                if (value.toString() !== params[key].toString()) matched = false;
            });
            return matched;
        }

        const check_for_matches = widget => {
            const name = widget.options.task.name();
            const params = get_actual_values(widget);

            let matched = null;
            demo_jobs.forEach(job => {
                if (job.name === name && params_match(job, params)) 
                    matched = job.job
            });

            return matched;
        }

        const intercept_run_button = widget => {
            if (widget.orig_submit) return;                         // Skip if already intercepted
            widget.orig_submit = widget.submit;                     // Store the original submit method
            widget.submit = () => {                                 // Replace with this method
                let matched_job_number = check_for_matches(widget); // Does a mock exist for this job?
                if (matched_job_number) {                           // If there's a match, show the job!
                    widget.expandCollapse(false);                   // And collapse the task widget
                    widget.add_job_widget(widget.options.cell, widget.options.session_index, matched_job_number);
                    let job_widget = widget.options.cell.element.find(".gp-widget-job").data("widget");
                    job_widget.infoMessage("This job contains pre-computed results used to demo GenePattern functionality.");
                    Jupyter.notebook.kernel.execute(`job${matched_job_number} = gp.GPJob(genepattern.session.get(0), ${matched_job_number})`, {}, {});
                }
                else {                                              // If not, ask the user
                    if (confirm('You are running in demo mode and no demo job can be found which ' + 
                                'match your inputs.\\n\\nWould you like to submit the job anyway?')) {
                        widget.orig_submit();                       // Submit the job the normal way
                    }
                }
            }
        }

        const demo_mode_on = () => {
            $(".gp-widget-task").each((i, task_node) => {
                const widget = $(task_node).data("widget");
                intercept_run_button(widget);
            });
        }

        setInterval(demo_mode_on, 1000);
    """))
